//
// Project : webgen
// File    : main.cc
// Purpose : main-function
// License : GPL
//
// arch-tag: 259c11ec-c8b4-4a9d-b800-2c474635d332
//

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <getopt.h>

#include <fstream>
#include <map>

#include "TSLL.hh"
#include "Menu.hh"
#include "parser.hh"

extern char * optarg;
extern int    optind, optopt;

using std::cout;
using std::endl;
using std::ostream;
using std::ofstream;
using std::ifstream;
using std::string;
using std::map;

// forward decl.
static void   parse_cmd     ( int argc, char ** argv );
static void   usage         ( int status, const string & param = "" );
static void   print_node    ( const string & id );
static void   print_node    ( Menu * menu );
static void   print_menu    ( ostream & os, Menu * node, Menu * outnode, string ofs );
static void   print_tabs    ( ostream & os, Menu * node, Menu * outnode, string ofs );
static void   print_content ( ostream & os, Menu * node );
static void   print_header  ( ostream & os, Menu * node );
static string rel_link      ( const string & file1, const string & file2 );
static string basedir       ( string & file );
static void   adjust_sub    ( Menu * father, Menu * sub );
static void   adjust_menu   ( Menu * menu );
static bool   is_modified   ( Menu * node );

// map with all menu entries
map< string, Menu * > nodes;

// file containing menu description
static string menu_file   = "menu.desc";
static string header_file = "";
static string footer_file = "";
static string impressum   = "";

// nodes to print out
static TSLL< string > outnodes;

// verbose mode
static int verbosity      = 1;

// different supported languages
enum { LN_ENGLISH, LN_GERMAN };

// language to use
static int  language      = LN_ENGLISH;

// force update
static bool force         = false;

// version of webgen
static string VERSION     = "v0.5.1";

//
// check verbosity level
//
bool
verbose ( int v )
{
    return (v <= verbosity);
}

//
// main function
//
int
main ( int argc, char ** argv )
{
    Menu  root( NULL, "root", "root", "", "" );

    nodes[ "root" ] = & root;

    parse_cmd( argc, argv );

    parse( & root, menu_file );
    adjust_menu( & root );
    
    if (verbose(3))
        root.print();

    // print all wished nodes
    if ( outnodes.size() == 0 )
    {
        TSLL< Menu * >             node_list;
        TSLL< Menu * >::TIterator  iter;

        node_list.append( & root );
        while ( node_list.size() > 0 )
        {
            root.clear_marks();
                
            if ( node_list.first_item()->name() != "root" )
                print_node( node_list.first_item() );
            
            iter = node_list.first_item()->submenus().first();
            while ( ! iter.eol() )
            {
                node_list.append( iter() );
                ++iter;
            }// while

            node_list.remove_first();
        }// while
    }// if
    else
    {
        TSLL< string >::TIterator  iter = outnodes.first();

        while ( ! iter.eol() )
        {
            root.clear_marks();
            print_node( iter() );
            ++iter;
        }// while
    }// else
}

// options descriptor
static struct option longopts[] = {
    { "deutsch",    no_argument,       0, 'd' },
    { "footer",     required_argument, 0, '1' },
    { "force",      no_argument,       0, 'f' },
    { "help",       no_argument,       0, 'h' },
    { "header",     required_argument, 0, '0' },
    { "impressum",  required_argument, 0, 'i' },
    { "menu",       required_argument, 0, 'm' },
    { "silent",     no_argument,       0, 's' },
    { "verbose",    no_argument,       0, 'v' },
    { 0,            0,                 0, 0 }
};

//
// parse cmdline
//
static void
parse_cmd ( int argc, char ** argv )
{
    char  c;
    
    while (( c = getopt_long( argc, argv, "adfhi:m:sv", longopts, NULL )) != -1 )
    {
        switch ( c )
        {
        case '0' : header_file = optarg; break;
        case '1' : footer_file = optarg; break;
        case 'd' : language = LN_GERMAN; break;
        case 'f' : force = true; break;
        case 'h' : usage( 0 ); break;
        case 'i' : impressum = optarg; break;
        case 'm' : menu_file = optarg; break;
        case 's' : verbosity = 0; break;
        case 'v' : verbosity++; break;
        case ':' :
        case '?' :
        default :
            usage( 1, argv[optopt] );
        }// switch
    }// while

    if ( optind < argc )
    {
        while ( optind < argc )
            outnodes.append( argv[optind++] );
    }// if
}

//
// print usage info
//
static void
usage ( int status, const string & param )
{
    if ( status != 0 )
    {
        if ( language == LN_GERMAN )
        {
            if (param != "")
                cout << "falscher oder fehlender parameter (" << param << "), siehe -h" << endl;
            else
                cout << "falscher oder fehlender parameter, siehe -h" << endl;
        }// if
        else
        {
            if (param != "")
                cout << "wrong or missing parameter (" << param << "), try -h" << endl;
            else
                cout << "wrong or missing parameter, try -h" << endl;
        }// else
            
        exit( status );
    }// if

    if ( language == LN_GERMAN )
    {
        cout << "webgen " << VERSION << " von Ronald Kriemann" << endl
             << "Aufruf : webgen [ options ] node1 node2 ..." << endl
             << "mit folgenden Optionen :" << endl
             << "    -d | --deutsch          : benutze deutsch als sprache in den html-dateien" << endl
             << "    -f | --force            : erzwinge generierung von html-dateien" << endl
             << "         --footer <datei>   : benutze <datei> als abschluss in html-datei (standard=intern)" << endl
             << "    -h | --help             : zeige diese hilfe anweisungen" << endl
             << "         --header <datei>   : benutze <datei> als kopf in html-datei (standard=intern)" << endl
             << "    -i | --impressum <link> : benutze <link> als impressum (standard=aus men-def.)" << endl
             << "    -m | --menu <datei>     : benutze <datei> als men-definition (standard=menu.desc)" << endl
             << "    -s | --silent           : keine ausgaben" << endl
             << "    -v | --verbose          : erhhe ausgabestufe (default=1)" << endl;
    }// if
    else
    {
        cout << "webgen " << VERSION << " by Ronald Kriemann" << endl
             << "usage : webgen [ options ] node1 node2 ..." << endl
             << "where options include :" << endl
             << "    -d | --deutsch          : use german language in written html-files" << endl
             << "    -f | --force            : force generation of node(s)" << endl
             << "         --footer <file>    : use footer given in <file> (default=none)" << endl
             << "    -h | --help             : print this info" << endl
             << "         --header <file>    : use header given in <file> (default=none)" << endl
             << "    -i | --impressum <link> : use <link> as impressum (default=from menu-desc)" << endl
             << "    -m | --menu <file>      : use menu in <file>" << endl
             << "    -s | --silent           : be silent" << endl
             << "    -v | --verbose          : increment verbosity level (default=1)" << endl;
    }// else
    
    exit( status );
}

//
// adjust menu entries, copy data from father to sons/tabs
//
static void
adjust_sub ( Menu * father, Menu * sub )
{
    TSLL< std::string >::TIterator  iter;
    
    // inherit title, etc. if not set
    if ( sub->logo()      == "" ) sub->logo()      = father->logo();
    if ( sub->title()     == "" ) sub->title()     = father->title() + " : " + sub->name();
    if ( sub->headline()  == "" ) sub->headline()  = father->headline();
    if ( sub->header()    == "" ) sub->header()    = father->header();
    if ( sub->footer()    == "" ) sub->footer()    = father->footer();
    if ( sub->impressum() == "" ) sub->impressum() = father->impressum();
    
    // prepend css-files of father to sons
    if ( ! sub->with_new_css() )
    {
        TSLL< std::string >  tmp;
        
        iter = father->css_files().first();

        // copy to temporary list to preserve order
        while ( ! iter.eol() )
        {
            tmp.prepend( iter() );
            ++iter;
        }// while

        while ( tmp.size() > 0 )
            sub->css_files().prepend( tmp.behead() );
    }// if

    if (verbose(4))
    {
        cout << "node = (" << sub->name() << ")" << endl
             << "  father    = (" << sub->father()->name() << ")" << endl
             << "  content   = (" << sub->content() << ")" << endl
             << "  html_file = (" << sub->html_file() << ")" << endl
             << "  logo      = (" << sub->logo() << ")" << endl
             << "  title     = (" << sub->title() << ")" << endl
             << "  headline  = (" << sub->headline() << ")" << endl
             << "  header    = (" << sub->header() << ")" << endl
             << "  footer    = (" << sub->footer() << ")" << endl
             << "  impressum = (" << sub->impressum() << ")" << endl;
        
        if ( sub->css_files().size() > 0 )
        {
            cout << "  cssfiles  = (";
            
            iter = sub->css_files().first();
            
            while ( ! iter.eol() > 0 )
            {
                cout <<  iter() << ", ";
                ++iter;
            }// while
            
            cout << ")" << endl;
        }// if

        if ( sub->is_hidden() )
            cout << "  hidden" << endl;
    }// if
}

static void
adjust_menu ( Menu * menu )
{
    TSLL< Menu * >::TIterator   iter = menu->submenus().first();
    
    while ( ! iter.eol() )
    {
        adjust_sub( menu, iter() );
        
        // recursive call 
        adjust_menu( iter() );
        
        ++iter;
    }// while

    iter = menu->tabs().first();

    while ( ! iter.eol() )
    {
        adjust_sub( menu, iter() );
        
        // recursive call 
        adjust_menu( iter() );
        
        ++iter;
    }// while
}

//
// write date into string
//
static bool
is_file ( std::string & file )
{
    struct stat  sb;

    if ( stat( file.c_str(), & sb ) == -1 )
        return false;

    return true;
}

//
// write date into string
//
static std::string
getdate ( Menu * node )
{
    struct tm  * timeval;
    char         buf[256];
    struct stat  sb;
    time_t       mtime;

    if ( stat( node->content().c_str(), & sb ) == -1 )
        mtime = time( NULL );
    else
        mtime = sb.st_mtime;
    
    timeval = localtime( & mtime );
    
    snprintf( buf, sizeof(buf), "%04d-%02d-%02d",
              timeval->tm_year + 1900, timeval->tm_mon+1, timeval->tm_mday );
    
    return std::string( buf );
}

//
// print single node
//
static void
print_node ( const string & id )
{
    if ( id == "" )
        return;
    
    //
    // go through menu and mark all fathers of node
    // and their direct submenus
    //

    Menu * node = nodes[ id ];

    if ( node == NULL )
    {
        if ( language == LN_GERMAN )
            cout << "FEHLER : kann knoten mit html-datei \"" << id << "\" nicht finden" << endl;
        else
            cout << "ERROR : could not find node with html-file = " << id << endl;
        return;
    }// if

    print_node( node );
}

static void
print_node ( Menu * node )
{
    if ( node == NULL )
        return;
    
    //
    // go through menu and mark all fathers of node
    // and their direct submenus
    //

    if ( verbose(2) )
    {
        if ( language == LN_GERMAN )
            cout << "drucke knoten " << node->name() << " ";
        else
            cout << "printing node " << node->name() << " ";
        
        if ( node->is_link() )
            cout << "(link)" << endl;
        else
            cout << "(" << node->html_file() << ")" << endl;
    }// if

    if ( node->is_link() )
        return;

    //
    // check if node has tabs and print them instead
    //
    
    if ( ! node->is_uniq() )
    {
        TSLL< Menu * >::TIterator  iter = node->tabs().first();

        while ( ! iter.eol() )
        {
            print_node( iter()->html_file() );
            ++iter;
        }// while

        return;
    }// if
    
    Menu                     * father;
    TSLL< Menu * >::TIterator  iter;

    node->set_mark( 1 );
    father = node;
    
    while ( true )
    {
        if ( father == NULL )
            break;
        
        father->set_mark( 1 );
        father = father->father();
    }// while

    //
    // check if dependencies were changed
    //

    if ( ! force && ! is_modified( node ) )
        return;

    if ( verbose(1) && ! verbose(2) )
    {
        if ( language == LN_GERMAN )
            std::cout << "generiere " << node->html_file() << endl;
        else
            std::cout << "generating " << node->html_file() << endl;
    }// if
    
    //
    // now that we have marked all nodes in the menu, which
    // should be printed in the table, let's do it already
    //

    ofstream  file( node->html_file().c_str() );

    if ( ! file )
    {
        if ( language == LN_GERMAN )
            cout << "FEHLER : kann datei \"" << node->html_file() << "\" nicht ffnen" << endl;
        else
            cout << "ERROR : could not open output-file = " << node->html_file() << endl;
        return;
    }// if

    // first the header of the html-file
    string  cssfile;

    file << "<?xml version=\"1.0\"?>" << endl
         << "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.1//EN\"" << endl
         << "    \"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd\">" << endl
         << endl
         << "<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\">" << endl
         << endl
         << "<head>" << endl
         << "  <title> " << node->title() << " </title>" << endl
         << "  <meta http-equiv=\"Content-Type\" content=\"text/html; charset=iso-8859-1\" />" << endl;
    
    file << "  <style type=\"text/css\">" << endl;
    
    TSLL< std::string >::TIterator  iter_css = node->css_files().first();
    
    while ( ! iter_css.eol() )
    {
        file << "  @import url(" <<
            rel_link( iter_css(), node->html_file() ) << ") screen;" << endl;
        ++iter_css;
    }// while

    file << "  @import url("
         << rel_link( "css/print.css", node->html_file() ) << ") print;" << endl;
    file << "</style>" << endl;
        
    file << "</head>" << endl
         << endl
         << "<body>" << endl
         << endl;

    file << "<div id=\"page\">" << endl;

    print_header( file, node );
        
    file << "<div id=\"menubox\">" << endl
         << "  <div id=\"menu\">" << endl
         << "    <ul>" << endl;

    print_menu( file, nodes[ "root" ], node, "    " );

    file << "    </ul>" << endl
         << "  </div>" << endl
         << "</div>" << endl;

    file << "<div id=\"contentbox\">" << endl;

    if ( node->is_tab() )
    {
        file << endl
             << "  <div id=\"tabs\">" << endl;
            
        print_tabs( file, nodes[ "root" ], node, "    " );
            
        file << "  </div>" << endl
             << endl;
    }// if

    file << "  <div id=\"content\">" << endl;
        
    // print the content
    print_content( file, node );

    file << "  </div>" << endl
         << "</div>" << endl;

    file << "<div id=\"footerbox\">" << endl
         << "  <div id=\"footer\">" << endl
         << "    <div id=\"modified\">" << endl;
    if ( language == LN_ENGLISH )
    {
        file << "      This page was last modified " << getdate( node ) << " by" << endl
             << "      <a href=\"mailto:rok@mis.mpg.de\"> Ronald Kriemann </a>" << endl;
    }// if
    if ( language == LN_GERMAN )
    {
        file << "      Diese seite wurde zuletzt am " << getdate( node ) << " von" << endl
             << "      <a href=\"mailto:rok@mis.mpg.de\"> Ronald Kriemann </a> gendert." << endl;
    }// if
    file << "    </div>" << endl;
    
    std::string  impstr = node->impressum();

    if ( impressum != "" )
        impstr = impressum;
    
    if ( impstr != "" )
    {
        file << "    <div id=\"impressum\">" << endl;

        if ( is_file( impstr ) )
            impstr = rel_link( impstr, node->html_file() );
        
        if ( language == LN_ENGLISH )
            file << "      <a href=\"" << impstr << "\">Contact</a>" << endl;

        if ( language == LN_GERMAN )
            file << "      <a href=\"" << impstr << "\">Impressum</a>" << endl;
        
        file << "    </div>" << endl;
    }// if
    
    file << "    <div id=\"w3check\">" << endl
         << "      <a href=\"http://validator.w3.org/check/referer\">" << endl;
    if ( language == LN_ENGLISH )
        file << "      Best viewed with any browser</a>" << endl;
    if ( language == LN_GERMAN )
        file << "      Darstellbar mit jedem browser</a>" << endl;
    file << "    </div>" << endl
         << "  </div>" << endl
         << "</div>" << endl;

    file << "</div>" << endl
         << endl
         << "</body>" << endl
         << "</html>" << endl;
}

//
// print menu to html-file
//
static void
print_menu ( ostream & os, Menu * node, Menu * outnode, string ofs )
{
    TSLL< Menu * >::TIterator  iter = node->submenus().first();
    Menu                     * father = NULL;

    if ( outnode->is_tab() )
        father = outnode->father();
        
    while ( ! iter.eol() )
    {
        if (( iter() == outnode ) || (( father != NULL ) && ( iter() == father )))
            os << ofs << "  <li class=\"active\"> " << iter()->name();
        else if ( ! iter()->is_hidden() )
        {
            os << ofs << "  <li>"
               << "<a href=\"";

            if ( iter()->is_link() )
                os << iter()->html_file() << "\">";
            else
            {
                if ( ! iter()->is_uniq() )
                    os << rel_link( iter()->tabs().first_item()->html_file(),
                                    outnode->html_file() )  << "\">";
                else
                    os << rel_link( iter()->html_file(),
                                    outnode->html_file() )  << "\">";
            }// else
                    
            os << iter()->name() << "</a>";
        }// else
           
        if ( ! iter()->is_hidden() || ( iter() == outnode ))
        {
            if ( ! ((iter()->mark() == 0) || (iter()->nsubmenus() == 0)) )
            {
                os << ofs << "  <ul>" << endl;
            
                print_menu( os, iter(), outnode, ofs + "  " );

                os << ofs << "  </ul>" << endl;
                os << ofs << "  </li>" << endl;
            }
            else
                os << "  </li>" << endl;
        }// if
            
        ++iter;
    }// while
}

//
// print tabs to html-file
//
static void
print_tabs ( ostream & os, Menu * node, Menu * outnode, string ofs )
{
    Menu * father = outnode->father();

    if ( father == NULL )
    {
        if ( language == LN_GERMAN )
            cout << "FEHLER: tab ohne vater" << endl;
        else
            cout << "ERROR: tab without father" << endl;
        return;
    }// if
        
    TSLL< Menu * >::TIterator  iter = father->tabs().first();
        
    while ( ! iter.eol() )
    {
        if ( iter() == outnode )
            os << ofs << "<div class=\"active\">" << iter()->name() << "</div>" << endl;
        else
            os << ofs << "<a href=\""
               << rel_link( iter()->html_file(),
                            outnode->html_file() )  << "\">"
               << iter()->name() << "</a>" << endl;
            
        ++iter;
    }// while
}

//
// print the content of a node into stream
//
static void
print_content ( ostream & os, Menu * node )
{
    ifstream  is( node->content().c_str() );

    if ( ! is )
    {
        if ( language == LN_GERMAN )
            cout << "FEHLER : kann datei \"" << node->content() << "\" nicht ffnen" << endl;
        else
            cout << "ERROR : could not open file (" << node->content() << ")" << endl;
        return;
    }// if

    char ch;

    while ( is.get(ch) )
        os.put( ch );
}

//
// print header
//
static void
print_header ( ostream & os, Menu * node )
{
    string    file = (header_file == "" ? node->header() : header_file);
    ifstream  is( file.c_str() );

    if ( (file == "") || ! is )
    {
        if ( node->headline() != "" )
        {
            os << "<div id=\"header\">" << endl
               << "  <div id=\"logobox\"><div id=\"logo\">" << node->logo() << "</div></div>" << endl
               << "  <div id=\"titlebox\"><div id=\"title\">" << node->headline() << "</div></div>" << endl
               << "</div>" << endl;
        }// if
    
        return;
    }// if

    char ch;
    
    while ( is.get(ch) )
        os.put( ch );
}

//
// return relative link from node2 to node1
//
static string
rel_link ( const string & file1, const string & file2 )
{
    string  link1 = file1;
    string  link2 = file2;
    string  link, d1, d2;

    // cut off common parent directories
    while ( true )
    {
        d1 = basedir( link1 );
        d2 = basedir( link2 );

        if ( d1 != d2 )
        {
            link1 = d1 + link1;
            link2 = d2 + link2;
            break;
        }// if

        if ( d1 == "" )
            break;
    }// while

    while ( basedir( link2 ) != "" )
        link += "../";

    link += link1;

    return link;
}

//
// return name of first parent directory
//
static string
basedir ( string & file )
{
    string    dir;
    unsigned  i;

    for ( i = 0; i < file.length(); i++ )
    {
        if ( file[i] == '/' )
        {
            for ( unsigned j = 0; j <= i; j++ )
                dir += file[j];

            file = & file.c_str()[i+1];

            return dir;
        }// if
    }// for

    return "";
}

//
// return true if html-file has to be generated
//
static bool
is_modified ( Menu * node )
{
    if ( node == NULL )
        return false;
    
    //
    // check content file and menu desc
    //

    struct stat  sb;
    double       content_time, html_time, menu_time;

    if ( stat( node->html_file().c_str(), & sb ) == -1 )
        return true;

    html_time = sb.st_mtime;
    
    if ( stat( node->content().c_str(), & sb ) == -1 )
        return true;

    content_time = sb.st_mtime;
    
    if ( stat( menu_file.c_str(), & sb ) == -1 )
        return true;

    menu_time = sb.st_mtime;

    if ( content_time > html_time )
    {
        if ( verbose(2) )
        {
            if ( language == LN_GERMAN )
                std::cout << node->content() << " ist neuer als " << node->html_file() << std::endl;
            else
                std::cout << node->content() << " is newer than " << node->html_file() << std::endl;
        }// if
        return true;
    }// if
    
    if ( menu_time > html_time )
    {
        if ( verbose(2) )
        {
            if ( language == LN_GERMAN )
                std::cout << menu_file << " ist neuer als " << node->html_file() << std::endl;
            else
                std::cout << menu_file << " is newer than " << node->html_file() << std::endl;
        }// if                
        return true;
    }// if
    
    return false;
}
